@@ -1,3 +1,7 @@
+"""
+Main package for texasholdem
+"""
+
 from texasholdem.game.game import TexasHoldEm
 from texasholdem.game.hand_phase import HandPhase
 from texasholdem.game.action_type import ActionType
